from types import SimpleNamespace

FEATURE_ID_VAR_DEFAULT = "feature_id"

RESOLVE_MATCHES_AGGREGATORS = SimpleNamespace(
    WEIGHTED_MEAN="weighted_mean", MEAN="mean", FIRST="first", MAX="max"
)

RESOLVE_MATCHES_TMP_WEIGHT_COL = "__tmp_weight_for_aggregation__"

BIND_DICT_OF_WIDE_RESULTS_STRATEGIES = SimpleNamespace(
    CONTATENATE="concatenate", MULTIPLE_KEYS="multiple_keys", STAGGER="stagger"
)
BIND_DICT_OF_WIDE_RESULTS_STRATEGIES_LIST = [
    BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.CONTATENATE,
    BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.MULTIPLE_KEYS,
    BIND_DICT_OF_WIDE_RESULTS_STRATEGIES.STAGGER,
]
